#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <bios.h>
#include <dos.h>
#include <conio.h>


#define BASE_PORT1	0x2a20
#define BASE_PORT2	0x2a30


#define _BV(x)		(1 << (x))

#define PIC_REG_EOI	0x20

#define REG_CNTR0	0x4
#define REG_CNTR1	0x5
#define REG_CNTR2	0x6
#define REG_TCWR	0x7
#define REG_TCR		0x8
#define REG_TSR		0xc


#define TCR_IBE		_BV(7)
#define TCR_TMSK	_BV(6)
#define TCR_EXT8	_BV(4)
#define TCR_TBE		_BV(3)
#define TCR_TAE		_BV(2)
#define TCR_TBC		_BV(1)
#define TCR_TAC		_BV(0)

#define TSR_TCS		_BV(7)
#define TSR_TBS		_BV(1)
#define TSR_TAS		_BV(0)


static int tcr = 0;


int get_tsr(unsigned port_base)
{
	return inp(port_base + REG_TSR);
}

void tcr_reset(unsigned port_base)
{
	tcr = 0;
	outp(port_base + REG_TCR, tcr);
}

void tcr_set(unsigned port_base, int val)
{
	tcr |= val;
	outp(port_base + REG_TCR, tcr);
//	printf("TCR=0x%x\r\n", tcr);
}

void tcr_clear(unsigned port_base, int val)
{
	tcr &= ~val;
	outp(port_base + REG_TCR, tcr);
//	printf("TCR=0x%x\r\n", tcr);
}

void tcr_rearm(unsigned port_base, int val)
{
	tcr_clear(port_base, val);
	tcr_set(port_base, val);
}


// timer
unsigned timer_get(unsigned port_base, int timer)
{
	unsigned val;
	switch(timer) {
	case 0:
		// latches contents of counter 0 to the register
		outp(port_base + REG_TCWR, 0x00);
		val = inp(port_base + REG_CNTR0);
		val = val | (inp(port_base + REG_CNTR0) << 8);
		break;
	case 1:
		// latches contents of counter 1 to the register
		outp(port_base + REG_TCWR, 0x40);
		val = inp(port_base + REG_CNTR1);
//		val = val | (inp(port_base + REG_CNTR1) << 8);
		break;
	case 2:
		// latches contents of counter 2 to the register
		outp(port_base + REG_TCWR, 0x80);
		val = inp(port_base + REG_CNTR2);
//		val = val | (inp(port_base + REG_CNTR2) << 8);
		break;
	default:
		val = 0;
	}

	return val;
}

void timer_set(unsigned port_base, int timer, unsigned val)
{
	switch(timer) {
	case 0:
		// latches contents of counter 0 to the register
		outp(port_base + REG_TCWR, 0x34);
		outp(port_base + REG_CNTR0, val & 0xff);
		outp(port_base + REG_CNTR0, val >> 8);
		break;
	case 1:
		// latches contents of counter 1 to the register
		outp(port_base + REG_TCWR, 0x74);
		outp(port_base + REG_CNTR1, val & 0xff);
		outp(port_base + REG_CNTR1, val >> 8);
		break;
	case 2:
		// latches contents of counter 2 to the register
		outp(port_base + REG_TCWR, 0xb6);
		outp(port_base + REG_CNTR2, val & 0xff);
		outp(port_base + REG_CNTR2, val >> 8);
		break;
	}
}

void timer_b_poll(void)
{
	int tsr;
	int i;

	tcr_set(BASE_PORT1, TCR_TBE | TCR_TBC);

	timer_set(BASE_PORT1, 1, 200);
	timer_set(BASE_PORT1, 2, 10000);

	for (i = 0; i < 160; i++) {
		tsr = get_tsr(BASE_PORT1);
		if (tsr & TSR_TBS) {
			tcr_rearm(BASE_PORT1, TCR_TBC);
			printf("Poll B\r\n");
		}
		delay(100);
	}
}


// timer interrupt test

#define IMFC_INT	0x0d

static unsigned int_cnt = 0;
void interrupt (*dos_isr)();

void interrupt imfc_isr()
{
	int tsr;

	_disable();

	// mask IMFC interrupt
	tcr_clear(BASE_PORT1, TCR_TMSK);


	// accept interrupt
	outp(PIC_REG_EOI, 0x20);

	_enable();

	tsr = get_tsr(BASE_PORT1);
	if (tsr & TSR_TBS) {
		int_cnt++;
		tcr_rearm(BASE_PORT1, TCR_TBC);
	}

	tcr_set(BASE_PORT1, TCR_TMSK);
}


void timer_b_int()
{
	int tsr;
	int i;
	unsigned old_int_cnt;

	printf("Timer B interrupt test\r\n");

	tsr = inp(0x21);
	printf("MSK=0x%x\r\n", tsr);
	outp(0x21, tsr & ~(1 << 5));

	// IRQ5 = INT_D
	_disable();

	dos_isr = _dos_getvect(IMFC_INT);
	_dos_setvect(IMFC_INT, imfc_isr);

	timer_set(BASE_PORT1, 1, 200);
	timer_set(BASE_PORT1, 2, 10000);

	tcr_set(BASE_PORT1, TCR_TBE | TCR_TBC);
	tcr_set(BASE_PORT1, TCR_TMSK | TCR_IBE);

	// unmask PIC1

	_enable();

	old_int_cnt = int_cnt;
	for (i = 0; i < 160; i++) {
//		tsr = get_tsr(BASE_PORT1);
//		printf("Int B counter %u, TSR=0x%x\r\n", int_cnt, tsr);
		if (int_cnt != old_int_cnt) {
			old_int_cnt = int_cnt;
			printf("Int B\r\n");
		}
		delay(100);
	}

	_disable();

	tsr = inp(0x21);
	printf("MSK=0x%x\r\n", tsr);
	outp(0x21, tsr | (1 << 5));

	tcr_clear(BASE_PORT1, TCR_TMSK | TCR_IBE);
	_dos_setvect(IMFC_INT, dos_isr);

	_enable();
}

int main(int argc, char *argv[])
{
	int i;
	unsigned c;
	int tsr;

	tcr_reset(BASE_PORT1);


//	timer_set(BASE_PORT1, 0, 32);

//	timer_set(BASE_PORT1, 2, 10000);
//	timer_set(BASE_PORT1, 1, 200);

//	timer_b_poll();
	timer_b_int();

	return 0;
}
